#include "language.h"

#include <string.h>
#include <share.h>
#include <fcntl.h>
#include <stdlib.h>
#include <time.h>
#include <builtin.h>

#include "misc.h"
#include "freq.h"
#include "session.h"

/*
  Type-2 Packet Format (proposed, but currently in use)
  -----------------------------------------------------
    Field    Ofs Siz Type  Description                Expected value(s)
    -------  --- --- ----  -------------------------- -----------------
    OrgNode  0x0   2 Word  Origination node address   0-65535
    DstNode    2   2 Word  Destination node address   1-65535
    Year       4   2  Int  Year packet generated      19??-2???
    Month      6   2  Int  Month  "        "          0-11 (0=Jan)
    Day        8   2  Int  Day    "        "          1-31
    Hour       A   2  Int  Hour   "        "          0-23
    Min        C   2  Int  Minute "        "          0-59
    Sec        E   2  Int  Second "        "          0-59
    Baud      10   2  Int  Baud Rate (not in use)     ????
    PktVer    12   2  Int  Packet Version             Always 2
    OrgNet    14   2 Word  Origination net address    1-65535
    DstNet    16   2 Word  Destination net address    1-65535
    PrdCodL   18   1 Byte  FTSC Product Code     (lo) 1-255
  * PVMajor   19   1 Byte  FTSC Product Rev   (major) 1-255
    Password  1A   8 Char  Packet password            A-Z,0-9
  * QOrgZone  22   2  Int  Orig Zone (ZMailQ,QMail)   1-65535
  * QDstZone  24   2  Int  Dest Zone (ZMailQ,QMail)   1-65535
    Filler    26   2 Byte  Spare Change               ?
| * CapValid  28   2 Word  CW Byte-Swapped Valid Copy BitField
  * PrdCodH   2A   1 Byte  FTSC Product Code     (hi) 1-255
  * PVMinor   2B   1 Byte  FTSC Product Rev   (minor) 1-255
  * CapWord   2C   2 Word  Capability Word            BitField
  * OrigZone  2E   2  Int  Origination Zone           1-65535
  * DestZone  30   2  Int  Destination Zone           1-65535
  * OrigPoint 32   2  Int  Origination Point          1-65535
  * DestPoint 34   2  Int  Destination Point          1-65535
  * ProdData  36   4 Long  Product-specific data      Whatever
    PktTerm   3A   2 Word  Packet terminator          0000

  * - extensions to FTS-0001

*/

#pragma pack(1)
struct PKTheader
{
  short origNode, destNode, year, month, day, hour, minute, second, baud;
  short level;
  short origNet, destNet;
  char  prodCode, subCode;
  char  password[8];
  short origZone, destZone;
  char  fill1[6];
  short capability;
  short orig1Zone, dest1Zone;
  short origPoint, destPoint;
  long  prtoductSpec;
  short pktTerm;
};
#pragma pack()

/*
                               Stored Message

       Offset
      dec hex
              .-----------------------------------------------.
        0   0 |                                               |
              ~                 fromUserName                  ~
              |                   36 bytes                    |
              +-----------------------+-----------------------+
       36  24 |                                               |
              ~                  toUserName                   ~
              |                   36 bytes                    |
              +-----------------------+-----------------------+
       72  48 |                                               |
              ~                    subject                    ~
              |                   72  bytes                   |
              +-----------------------+-----------------------+
      144  90 |                                               |
              ~                    DateTime                   ~
              |                    20 bytes                   |
              +-----------------------+-----------------------+
      164  A4 | timesRead (low order) | timesRead (high order)|
              +-----------------------+-----------------------+
      166  A6 | destNode (low order)  | destNode (high order) |
              +-----------------------+-----------------------+
      168  A8 | origNode (low order)  | origNode (high order) |
              +-----------------------+-----------------------+
      170  AA |   cost (low order)    |   cost (high order)   |
              +-----------------------+-----------------------+
      172  AC | origNet (low order)   | origNet (high order)  |
              +-----------------------+-----------------------+
      174  AE | destNet (low order)   | destNet (high order)  |
              +-----------------------+-----------------------+
      176  B0 | destZone (optional)   | destZone (optional)   |
              +-----------------------+-----------------------+
      178  B2 | origZone (optional)   | origZone (optional)   |
              +-----------------------+-----------------------+
      180  B4 | destPoint(optional)   | destPoint(optional)   |
              +-----------------------+-----------------------+
      182  B6 | origPoint(optional)   | origPoint(optional)   |
              +-----------------------+-----------------------+
      184  B8 |  replyTo (low order)  |  replyTo (high order) |
              +-----------------------+-----------------------+
      186  BA | Attribute (low order) | Attribute (high order)|
              +-----------------------+-----------------------+
      188  BC | nextReply (low order) | nextReply (high order)|
              +-----------------------+-----------------------+
      190  BE |                      text                     |
              ~                    unbounded                  ~
              |                 null terminated               |
              `-----------------------------------------------'

      Message    = fromUserName(36)  (* Null terminated *)
                   toUserName(36)    (* Null terminated *)
                   subject(72)       (* see FileList below *)
                   DateTime          (* message body was last edited *)
                   timesRead         (* number of times msg has been read *)
                   destNode          (* of message *)
                   origNode          (* of message *)
                   cost              (* in lowest unit of originator's
                                        currency *)
                   origNet           (* of message *)
                   destNet           (* of message *)
                   destZone          (* of message *)
                   origZone          (* of message *)
                   destPoint         (* of message *)
                   origPoint         (* of message *)
                   replyTo           (* msg to which this replies *)
                   AttributeWord
                   nextReply         (* msg which replies to this *)
                   text(unbounded)   (* Null terminated *)

      DateTime   = (* a character string 20 characters long *)
                                     (* 01 Jan 86  02:34:56 *)
                   DayOfMonth " " Month " " Year " "
                   " " HH ":" MM ":" SS
                   Null

      DayOfMonth = "01" | "02" | "03" | ... | "31"   (* Fido 0 fills *)
      Month      = "Jan" | "Feb" | "Mar" | "Apr" | "May" | "Jun" |
                   "Jul" | "Aug" | "Sep" | "Oct" | "Nov" | "Dec"
      Year       = "01" | "02" | .. | "85" | "86" | ... | "99" | "00"
      HH         = "00" | .. | "23"
      MM         = "00" | .. | "59"
      SS         = "00" | .. | "59"

      AttributeWord   bit       meaning
                      ---       --------------------
                        0  +    Private
                        1  + s  Crash
                        2       Recd
                        3       Sent
                        4  +    FileAttached
                        5       InTransit
                        6       Orphan
                        7       KillSent
                        8       Local
                        9    s  HoldForPickup
                       10  +    unused
                       11    s  FileRequest
                       12  + s  ReturnReceiptRequest
                       13  + s  IsReturnReceipt
                       14  + s  AuditRequest
                       15    s  FileUpdateReq

                             s - need not be recognized, but it's ok
                             + - not zeroed before packeting
*/

#pragma pack(1)
struct FIDO_msg
{
  short orig_node,
	dest_node,
	orig_net,
	dest_net;

  unsigned short attributes;
  short cost;
};
#pragma pack()

char * format( long freqSize, long freqTime,
			     long size, long time )
{
  static char ret[81];
  char lsize[32],
       ltime[32],
       usize[32],
       utime[32];

  if( freqSize )
    sprintf( lsize, Msg_InfoFREQSize, freqSize );
  else
    strcpy( lsize, Msg_InfoFREQWordNone );

  if( size )
    sprintf( usize, Msg_InfoFREQSize, size/1024 + (size%1024>511 ? 1 : 0) );
  else
    strcpy( usize, Msg_InfoFREQWordNone );

  if( freqTime )
    sprintf( ltime, Msg_InfoFREQTime, freqTime );
  else
    strcpy( ltime, Msg_InfoFREQWordNone );

  if( time )
    sprintf( utime, Msg_InfoFREQTime, time/60 + ( time%60>29 ? 1: 0) );
  else
    strcpy( utime, Msg_InfoFREQWordNone );

  strcpy( ret, lsize );
  strcat( ret, " " );
  strcat( ret, ltime );
  strcat( ret, " " );
  strcat( ret, usize );
  strcat( ret, " " );
  strcat( ret, utime );

  return ret;
}

void generateFreqAnswer( char *msgText, FIDOAddress& aka,
				char *hisName, char *hisPassword,
				void *fictionPointer )
{
  char *bink( aka.buildOutboundName( ".Rut", ".Request.Mail" ) );

  if( bink==NULL )
    return;

  char *name( tmpStr( bink ) );

  delete bink;

  makeFilePath( name );

  FILE *in( _fsopen( name, "wb" ) );

  if( in==NULL )
  {
    log( Msg_ErrUnableOpen, name, strError() );
    return;
  }

 unsigned short zone( defaultAddress->getZone() ),
		net( defaultAddress->getNet() ),
		node( defaultAddress->getNode() ),
		point( defaultAddress->getPoint() );

 char pntKludge1[32],
      pntKludge2[32],
      *intlKludge( NULL ),
      text[ 1024 ];

 unsigned short toZone( aka.getZone() ),
		toNet( aka.getNet() ),
		toNode( aka.getNode() ),
                toPoint( aka.getPoint() );

 if( point )
   sprintf( pntKludge1, "\01FMPT %u\r\n", point );

 if( toPoint )
   sprintf( pntKludge2, "\01TOPT %u\r\n", toPoint );

 char *klName( "\01INTL " ),
      address[65],
      toIntlAddr[65],
      fromIntlAddr[65];

 defaultAddress->toString( address, 65 );

 sprintf( fromIntlAddr, "%d:%d/%d", zone, net, node );
 sprintf( toIntlAddr, "%d:%d/%d", toZone, toNet, toNode );
 intlKludge = (char *)_alloca( strlen( fromIntlAddr )+
                               strlen( toIntlAddr )+
                               strlen( klName )+1+3 );

 strcpy( intlKludge, klName );
 strcat( intlKludge, toIntlAddr );
 strcat( intlKludge,  " " );
 strcat( intlKludge, fromIntlAddr );
 strcat( intlKludge,  "\r\n" );

 sprintf( text,
  "%s"
  "%s"
  "%s"
  "\01MSGID: %s %08lx%08lx\r\n"
  "\01PID: The-Brake! %s %s\r\n"
  "\01DOMAIN %s %d:%d/%d.%01d %s %d:%d/%d.%01d\r\n"
  "\01FLAGS PGM NPD\r\n",
  intlKludge,
  point ? pntKludge1 : "",
  toPoint ? pntKludge2 : "",
  address,
  time( NULL ),getTimeMark(),
  M_VERSION, M_NUMBER,
  aka.getDomain(), toZone, toNet, toNode, toPoint,
  defaultAddress->getDomain(), zone, net, node, point );

 PKTheader *pkt = (PKTheader *)_alloca( sizeof( PKTheader ) );

 memset( pkt, 0, sizeof(PKTheader) );

 time_t ttm = time( NULL );
 struct tm *tms = localTime( &ttm );

 pkt->origNode = node;
 pkt->destNode = toNode;
 pkt->year = short( tms->tm_year+1900 );
 pkt->month = short( tms->tm_mon+1 );
 pkt->day = short( tms->tm_mday );
 pkt->hour = short( tms->tm_hour );
 pkt->minute = short( tms->tm_min );
 pkt->second = short( tms->tm_sec );
 pkt->baud = short( comPort->getLastBaud() );
 pkt->level = 2;
 pkt->origNet = net;
 pkt->destNet = toNet;
 pkt->prodCode = 0xfe;
 pkt->subCode = 1;
 pkt->capability = 1;

 if( hisPassword!=NULL && illegalAka==NULL )
   strncpy( pkt->password, hisPassword, sizeof(pkt->password) );

 pkt->destZone = pkt->dest1Zone = toZone;
 pkt->origZone = pkt->orig1Zone = zone;
 pkt->origPoint = point;
 pkt->destPoint = toPoint;
 pkt->pktTerm = 2;

 FIDO_msg *msg = (FIDO_msg *)_alloca( sizeof( FIDO_msg ) );

 memset( msg, 0, sizeof(FIDO_msg) );

 char msg_date[32];

 sprintf( msg_date, "%02d %3s %02d  %02d:%02d:%02d",
  pkt->day, monthName[pkt->month-1], pkt->year%100,
  pkt->hour, pkt->minute, pkt->second );

 msg->orig_node=node;
 msg->dest_node=toNode;
 msg->cost=0;
 msg->orig_net=net;
 msg->dest_net=toNet;
 msg->attributes=0x2001;		// IsReturnReceipt+Private

 char null[] = { 0, 0, 0 },
      *origin = tmpStr( systemInfo.name ),
      oriStart[]( " * Origin: " ),
      oriEnd[ 65 ];

 int lenOri( strlen(origin) );

 defaultAddress->toString( oriEnd, 64 );

 if( lenOri+strlen(oriStart)+strlen(oriEnd)+3>78 )
   lenOri = 78 - 3 - strlen( oriEnd ) - strlen( oriStart );

 if( lenOri<0 )
   lenOri = 0;

 origin[ lenOri ] = 0;

 if( fwrite( pkt, 1, sizeof(PKTheader), in )!=sizeof(PKTheader) )
err_write:
   log( Msg_ErrUnableWrite, name, strError() );
 else
 {
   if( fwrite( msg, 1, sizeof(FIDO_msg), in )!=sizeof(FIDO_msg) )
     goto err_write;

   if( fwrite( msg_date, 1, 20, in )!=20 )
     goto err_write;

   if( fwrite( hisName, 1, strlen(hisName)+1, in )!=strlen(hisName)+1 )
     goto err_write;

   if( fwrite( M_NAME, 1, strlen(M_NAME)+1, in )!=strlen(M_NAME)+1 )
     goto err_write;

   if( illegalAka==NULL )
   {
     if( fwrite( Msg_InfoFREQmsgSubj, 1, strlen(Msg_InfoFREQmsgSubj)+1, in )!=strlen(Msg_InfoFREQmsgSubj)+1 )
       goto err_write;
   }
   else
     if( fwrite( Msg_InfoFREQmsgSubj1, 1, strlen(Msg_InfoFREQmsgSubj1)+1, in )!=strlen(Msg_InfoFREQmsgSubj1)+1 )
       goto err_write;

   if( fwrite( text, 1, strlen(text), in )!=strlen(text) )
     goto err_write;

   if( fprintf( in, "\r\n" )==EOF )
     goto err_write;

   if( fwrite( msgText, 1, strlen(msgText), in )!=strlen(msgText) )
     goto err_write;

   if( fictionPointer!=NULL )
   {
     if( fprintf( in, "\r\n" )==EOF )
       goto err_write;

     NodeFreqInfo *info = (NodeFreqInfo *)fictionPointer;
     char *fmt1 = "\r\n%-7s %-8s %-9s %-8s %-9s",
	  *fmt2 = "\r\n%-7s %s",
          buf[65];

     if( fprintf( in, Msg_InfoFREQLimitsFor, aka.toString( buf, 65 ) )==EOF )
       goto err_write;

     if( fprintf( in, "\r\n%7s %6c%s%15c%s", "", ' ', Msg_InfoFREQWLimit, ' ', Msg_InfoFREQWUsed )==EOF )
       goto err_write;

     if( fprintf( in, fmt1, "", Msg_InfoFREQWSize, Msg_InfoFREQWTime, Msg_InfoFREQWSize, Msg_InfoFREQWTime )==EOF )
       goto err_write;

     if( fprintf( in, fmt1, "", "--------", "---------", "--------", "---------" )==EOF )
       goto err_write;

     if( fprintf( in, fmt2, Msg_InfoFREQWSess,
		  format( miscInfo.sessionFreqSize, miscInfo.sessionFreqTime,
			  0, 0 ) )==EOF )
       goto err_write;

     if( fprintf( in, fmt2, Msg_InfoFREQWDay,
		  format( miscInfo.dayFreqSize, miscInfo.dayFreqTime,
			  info->daySize, info->dayTime ) )==EOF )
       goto err_write;

     if( fprintf( in, fmt2, Msg_InfoFREQWWeek,
		  format( miscInfo.weekFreqSize, miscInfo.weekFreqTime,
			  info->weekSize, info->weekTime ) )==EOF )
       goto err_write;

     if( fprintf( in, fmt2, Msg_InfoFREQWMonth,
		  format( miscInfo.monFreqSize, miscInfo.monFreqTime,
			  info->monSize, info->monTime ) )==EOF )
       goto err_write;

     if( fprintf( in, "\r\n\r\n" )==EOF )
       goto err_write;

     FILE *file = NULL;

     if( miscInfo.freqPolicy!=NULL )
       file = _fsopen( miscInfo.freqPolicy, "r" );

     if( file!=NULL )
     {
       while( 1 )
       {
	 int readed = getc( file );

	 if( readed==EOF )
	   break;
	 else
	   if( readed=='\n' )
	   {
	     if( putc( '\r', in )==EOF )
	       goto err_write;

	     if( putc( '\n', in )==EOF )
	       goto err_write;
	   }
	   else
	     if( putc( (char)readed, in )==EOF )
	       goto err_write;

       }

       fclose( file );
     }
   }

   if( fprintf( in, "\r\n\r\n--- %s %s/%s\r\n%s%s (%s)\r\n",
		M_NAME, M_VERSION, M_NUMBER,
		oriStart, origin, oriEnd )==EOF )
     goto err_write;

   if( fwrite( null, 1, sizeof(null), in )!=sizeof(null) )
     goto err_write;
 }

 fclose( in );
}
