#include <stdlib.h>
#include <memory.h>

#include "buffer.h"
#include "misc.h"

Buffer::Buffer( int c, int d ) :
  curLen( c ),
  pos( 0 ),
  delta( d )
{
  if( !delta )
    delta = 10;

  if( !curLen )
    curLen = delta;

  str = new char[ curLen ];
  reset();
}

Buffer::~Buffer()
{
  delete str;
}

void Buffer::putInto( char ch )
{
  if( pos==curLen )
  {
    char *old( str );

    str = new char [ curLen+delta ];

    memcpy( str, old, curLen );
    memset( str+curLen, 0, delta );

    curLen += delta;
    delete old;
  }

  str[ pos++ ] = ch;
}

void Buffer::putQuotted( unsigned ch )
{
  switch( ch )
  {
  case '\\':
    putInto( '\\' );
    putInto( '\\' );
     break;

  case '\r':
    putInto( '\\' );
    putInto( 'r' );
    break;

  case '\n':
    putInto( '\\' );
    putInto( 'n' );
    break;

  case '\b':
    putInto( '\\' );
    putInto( 'b' );
    break;

  case '\t':
    putInto( '\\' );
    putInto( 't' );
    break;

  case '\f':
    putInto( '\\' );
    putInto( 'f' );
    break;

  case ' ':
    putInto( '\\' );
    putInto( 's' );
    break;

  default:
    if( ch<0x20 || ch>0x7e )
    {
      char bbb[16];

      sprintf( bbb, "\\x%02x", ch&0xff );
      putString( bbb );
    }
    else
      putInto( char( ch ) );
  }
}

void Buffer::putQuotted( char *buf )
{
  while( *buf )
  {
    putQuotted( ((unsigned)(unsigned char)*buf)&0xff );
    buf++;
  }
}

void Buffer::put( register int ch )
{
  switch( ch )
  {
  case 0:
    putInto( 0 );
    break;

  case '\n':
    break;

  case '\r':
    putInto( '|' );
    break;

  default:
    if( ch<0x20 || ch>0x7f )
    {
      putInto( '\\' );
      putInto( 'x' );
      putInto( hexString[ (ch>>4)&0x0f ] );
      putInto( hexString[ ch&0x0f ] );

    }
    else
      putInto( char(ch) );
  }
}

void Buffer::putEMSI( register char *str )
{
  for( register char ch(0); ch=*str, ch; str++ )
    if( ch==']' || ch=='}' )
    {
      putInto( ch );
      putInto( ch );
    }
    else if( ch<0x20 || ch>0x7e || ch=='\\' )
    {
      putInto( '\\' );
      putInto( hexString[ (ch>>4)&0x0f ] );
      putInto( hexString[ ch&0x0f ] );
    }
    else
      putInto( ch );
}

/*
void Buffer::stripLider()
{
  register char strip(1);
  register int j(0);

  for( register int i(0); i<pos; i++ )
    if( strip )
    {
      if( str[i]!='|' )
      {
	strip = 0;
	str[j++] = str[i];
      }
    }
    else
      str[j++] = str[i];

  if( j!=pos )
  {
    memset( str+j, 0, pos-j );
    pos = j;
  }
}

void Buffer::stripTrailer()
{
  register int i( pos-2 );

  for( ; i>=0 && str[i]=='|'; i-- );

  if(  pos-i-1>0 )
    memset( str+i+1, 0, pos-i-1 );

  pos = i+2;
}
*/

void Buffer::reset()
{
  pos = 0;
  
  if( curLen )
    memset( str, 0, curLen );
  else
    *str = 0;
}

void Buffer::leftShift( int how )
{
  if( how>pos )
    return;
  else
    if( how!=pos )
      memcpy( str, str+how, pos-how );

  pos -= how;
  memset( str+pos, 0, curLen-pos );
}

int Buffer::operator[]( int index )
{
  if( index>=pos )
    return -1;
  else
    return int(str[index])&0xff;
}

void Buffer::putString( char *str )
{
  while( *str )
    putInto( *(str++) );
}

void Buffer::stripByte()
{
  if( pos )
    pos--;
}

void Buffer::deQuote()
{
  int cur(0),
      store(0);

  while( 1 )
  {
    int ch( getChar( cur ) );

    if( ch==EOF )
      break;

    str[ store++ ] = (char)ch;
  }

  cur = pos;
  pos = store;

  while( store < cur )
    str[ store++ ] = 0;
}

int Buffer::getChar( int& cur )
{
  char base( 10 ),
       state( 0 ),
       ctlChar( 0 ),
       left( 3 ),
       curDigit( 0 );

  register int ch;

  for( int done(0); !done; )
  {
    if( cur >= pos )
    {
      ch = EOF;
      break;
    }

    ch = str[ cur++ ];

    switch( state )
    {
    case 0:
      if( ch=='\\' )
	state = 1;
      else
        done = 1;

      break;

    case 1:
      switch( ch )
      {
      case '\n':
	state = 0;
	break;

      case 'r':
	ch = '\r';
	done = 1;
	break;

      case 'n':
	ch = '\n';
	done = 1;
	break;

      case 'b':
	ch = '\b';
	done = 1;
	break;

      case '\\':
	ch = '\\';
	done = 1;
	break;

      case 't':
	ch = '\t';
	done = 1;
	break;

      case 'f':
	ch = '\f';
	done = 1;
	break;

      case 's':
	ch = ' ';
	done = 1;
	break;

      case '0':
	base = 8;
	state = 2;
	break;

      case 'x':
      case 'X':
	base = 16;
	left = 2;
	state = 2;
	break;

      default:
	cur--;

	if( isDigit(ch) )
	  state = 2;
	else
	{
	  ch = '\\';
	  done = 1;
	}
      }

      break;

    case 2:
      if( isValidDigit( ch, base, curDigit, left ) )
	ctlChar = char( ctlChar*base + curDigit );
      else
      {
        cur--;

	ch = ctlChar;
	done = 1;
      }
      break;

    default:
      done = 1;
    }
  }

  return ch;
}

