#if !defined(__FIDO_ADDRESS)

#define __FIDO_ADDRESS

#include <stdio.h>

#include "class.h"
#include "worktime.h"
#include "nodelist.h"
#include "domain.h"

typedef unsigned short uShort;
typedef unsigned char uChar;
typedef unsigned long uLong;

#ifndef _SCHEDULE_MANIP
#define _SCHEDULE_MANIP
class Schedule;
void deleteSched( Schedule *sh, int delThis=1 );
Schedule * newSched( Schedule *init );
#endif

enum Flavour
{
  unknownFlavour=0, holdFlavour=1,
  normalFlavour=2, directFlavour=4,
  crashFlavour=8, immedFlavour=16,
  onlineFlavour=32, requestFlavour=64
};

int flavourPriority( Flavour );
char *flavourNameByPriority( int );
char *flavourString( Flavour );

extern int immedFlavourPriority;
extern int crashFlavourPriority;
extern int directFlavourPriority;
extern int normalFlavourPriority;
extern int holdFlavourPriority;

enum LongNameType
{
  unknownLongName=0,
  boxLongName,
  mailLongName,
  listLongName,
  requestLongName,
  busyLongName,
  holdLongName,
  tryLongName
};

struct Address
{
  uShort zone, net, node, point;
  char *domain;
};

enum FIDOAddressShowMode
{
  FIDOAddressShowModeCfg,
  FIDOAddressShowModeEMSI
};

class FIDOAddress : public BaseObject
{
  Address addr;
  char valid;
  Domain *dom;
  char unknowDomain;

  char *nameOfClass() const;

  void makeDomain();
  int createBusy( char *bsyFull );
  void deleteBusy( char *bsyFull );
  int makeLongBsy( int create );
  int makeBinkBsy( int create );
  char *buildListName( Flavour ) const;
  void removeFreqFile( char *bink ) const;
  static char *concatenateWithDot( const char *s1, const char *s2 );

public:

  FIDOAddress( uShort iZone, uShort iNet, uShort node,
	       uShort iPoint=0, char *iDomain=NULL, Domain *idom=NULL );
  FIDOAddress( char *s, Domain *idom=NULL );
  FIDOAddress( const FIDOAddress *a );
  FIDOAddress( const Address& a );
  FIDOAddress( const FIDOAddress & );
  virtual ~FIDOAddress();

  int equal( const BaseObject &t ) const;
  int less( const FIDOAddress &t ) const;
  void print( ostream& s ) const;

  static void updateLoName( Flavour flavourLO, char *xlo );

  char *buildOutboundName( char *sh, char *lo ) const;

  int is4D() const { return addr.point; }
  int is5D() const { return addr.domain!=NULL; }
  int isValid() const { return valid; }
  int isUnknowDomain() const { return unknowDomain; }

  char *getDomain() const { return addr.domain; }
  uShort getZone() const { return addr.zone; }
  uShort getNet() const { return addr.net; }
  uShort getNode() const { return addr.node; }
  uShort getPoint() const { return addr.point; }
  Domain *getDomainPtr() const { return dom; }

  Address *getAddressPointer() { return &this->addr; }

  int makeBsySemaphore( int );
  char *getBsySemaName() const;

  int makeBsy( int );
  int checkBsy() const;
  void makeLO( Flavour, int =1 ) const;
  void dropLO( Flavour ) const;
  void makeSend( char *, Flavour ) const;
  void makeFreq( char *, Flavour ) const;
  void makeHold( unsigned long ) const;
  void dropFreq() const;

  char *getTBoxName() const;
  char *getBinkName() const;
  char *getLongName() const;
  char *toString( char *buf, int len,
                  FIDOAddressShowMode mode = FIDOAddressShowModeCfg ) const;
  char *getLongBox() const;
  char *getLongBinkName() const;
  char *getBsyName() const;
  char *getLongBsyName() const;
  char *getTryFileName() const;
};

class FIDOAddressList : public Collection
{
public:
  FIDOAddressList();
  
  FIDOAddress* Find( FIDOAddress& item, LockMode lock=LockModeNeed ) const;
  FIDOAddress& operator [] ( const int pos ) const;
};

class FIDOAddressIterator : public CollectionIterator
{
public:
  FIDOAddressIterator( FIDOAddressList& ini );
  
  FIDOAddress& operator () () const;
};

class Hidden : public BaseObject
{
  int wait;
  char *phone,
       *dial,
       *suff;
  WorkTime *work;

  int equal( const BaseObject &t ) const;
  void print( ostream& s ) const;
  char *nameOfClass() const;

public:

  Hidden( int wait = 0, char *ph = NULL, char *dil = NULL,
          WorkTime *wt = NULL, char *su = NULL );
  Hidden( Hidden * );
  Hidden( Hidden & );
  virtual ~Hidden();

  int getWait() const { return wait; }
  char *getPhone() const { return phone; }
  char *getDial() const { return dial; }
  char *getDialSuff() const { return suff; }

  WorkTime *getWorkTime() const { return work; }

  int isWorked( long tt=0 ) const;
};

class HiddenList : public Collection
{
public:
  HiddenList();

  Hidden* Find( Hidden& item, LockMode lock=LockModeNeed ) const;
  Hidden& operator [] ( const int pos ) const;
};

class HiddenIterator : public CollectionIterator
{
public:
  HiddenIterator( HiddenList& ini );
  
  Hidden& operator () () const;
};

class NodeConfig : public FIDOAddress
{
  WorkTime *work;
  char *holdBox,
       *normalBox,
       *crashBox,
       *immedBox,
       *directBox,
       *inbound,
       *password,
       *phone,
       *dial,
       *suff,
       *flags,
       *alias;
  int wait;
  int baud;
  Flavour flavour;
  int priority;
  CacheItem *ndl;
  short maxIn,
	maxOut;
  int maxTry;
  FIDOAddressList *aka;
  HiddenList *hidden;
  int current;
  Schedule *after,
	   *before,
	   *doit;
  int noDZAenabled;
  int noZAPenabled;
  int noHydraEnabled;
  char *rexxDial;
  int cost;
  int chain;
  char *mask;

  char *nameOfClass() const;
  int equal( const BaseObject &t ) const;

protected:

  void init( char *ph=NULL, char *pass=NULL,
             char *norm=NULL, char *hold=NULL, char *crash=NULL,
             char *immed=NULL, char *direct=NULL,
             char *inbound=NULL,
             char *preff=NULL, int wt=0, char *fl=NULL,
             FIDOAddressList *list=NULL, HiddenList *lines=NULL,
             Schedule *before=NULL, Schedule *after=NULL,
             Schedule *doit=NULL, char *rexxDial=NULL, char *suff=NULL,
             char *alias=NULL );
public:

   NodeConfig( char *iAddr, char *iWork=NULL,
	      Flavour iF=unknownFlavour,
	      char *ph=NULL, char *pass=NULL,
	      char *norm=NULL, char *hold=NULL, char *crash=NULL,
              char *immed=NULL, char *direct=NULL,
	      char *inbound=NULL,
	      char *preff=NULL, int wt=0, char *fl=NULL,
	      short maxIn=0, short maxOut=0, int maxTry=0,
	      FIDOAddressList *list=NULL, HiddenList *lines=NULL,
	      Schedule *before=NULL, Schedule *after=NULL,
	      Schedule *doit=NULL, int noDZA=0, char *rexxDial=NULL,
              int cost=0, int baud=0, int noHydraEnabled=0, int chain=0,
              char *mask=NULL, int prio=0, char *suff=NULL, char *alias=NULL,
              int noZAP=0 );

   NodeConfig( FIDOAddress *iAddr, WorkTime *iWork=NULL,
	      Flavour iF=unknownFlavour,
	      char *ph=NULL, char *pass=NULL,
	      char *norm=NULL, char *hold=NULL, char *crash=NULL,
              char *immed=NULL, char *direct=NULL,
	      char *inbound=NULL,
	      char *preff=NULL, int wt=0, char *fl=NULL,
	      short maxIn=0, short maxOut=0, int maxTry=0,
	      FIDOAddressList *list=NULL, HiddenList *lines=NULL,
	      Schedule *before=NULL, Schedule *after=NULL,
	      Schedule *doit=NULL, int noDZA=0, char *rexxDial=NULL,
              int cost=0, int baud=0, int noHydraEnabled=0, int chain=0,
              char *mask=NULL, int prio=0, char *suff=NULL, char *alias=NULL,
              int noZAP=0 );

  NodeConfig( const NodeConfig *i );
  NodeConfig( const NodeConfig &i );

  virtual ~NodeConfig();

  void print( ostream& s ) const;

  void resolveFromMask();
  char *toString( char *buf, int len,
                  FIDOAddressShowMode mode = FIDOAddressShowModeCfg ) const;

  int isValid() const;
  int isWorked( uLong stamp=0 ) const;
  int isPrimaryWorked( uLong stamp=0 ) const;

  char *getImmedBox() const { return immedBox; }
  char *getDirectBox() const { return directBox; }
  char *getHoldBox() const { return holdBox; }
  char *getNormalBox() const { return normalBox; }
  char *getCrashBox() const { return crashBox; }
  char *getInbound() const { return inbound; }
  char *getPassword() const { return password; }
  char *getRexxDial() const { return rexxDial; }
  char *getAlias() const { return alias; }

  int getPriority() const { return priority; }

  char *getTranslatedPhone( char *orig );

  char *getPhone() const;
  char *getDial() const;
  char *getDialSuff() const;
  int getWaitCarrier() const;

  Flavour getCfgFlavour() const { return flavour; }
  int getCurrentLine() const { return current; }
  Hidden *getCurrentHidden() const;

  FIDOAddressList *getAkaList() const { return aka; }
  HiddenList *getHiddenList() const { return hidden; }

  CacheItem *getNdl() const { return ndl; }

  char *getFlags() const;
  long getBaud() const { return baud; }
  int getChain() const { return chain; }
  char *getMask() const { return mask; }

  short getMaxInComing() const;
  short getMaxOutGoing() const;
  int getMaxTry() const;
  int getCost() const { return cost; }

  int isMask() const { return mask!=NULL; }

  int haveImmedBox() const { return immedBox!=NULL; }
  int haveDirectBox() const { return directBox!=NULL; }
  int haveHoldBox() const { return holdBox!=NULL; }
  int haveNormalBox() const { return normalBox!=NULL; }
  int haveCrashBox() const { return crashBox!=NULL; }
  int haveInbound() const { return inbound!=NULL; }
  int havePassword() const { return password!=NULL; }

  int isListed() const { return ndl!=NULL; }

  void setFlavour( Flavour iState ) { flavour = iState; }

  void setCurrent();
  void findAvailable();

  int disableDZA() const { return noDZAenabled; }
  int disableZAP() const { return noZAPenabled; }
  int disableHydra() const { return noHydraEnabled; }

  Schedule *getBefore() const { return before; }
  Schedule *getAfter() const { return after; }
  Schedule *getDo() const { return doit; }

  WorkTime *getWorkTime() const { return work; }

  void changeAka( FIDOAddressList *list, int line );
  void changeHidden( HiddenList *hidden, int line );
  void changeWork( WorkTime *iwork );
  void changeMask( char *str );
  void changeImmedBox( char *str );
  void changeDirectBox( char *str );
  void changeHoldBox( char *str );
  void changeNormalBox( char *str );
  void changeCrashBox( char *str );
  void changeInbound( char *str );
  void changePassword( char *str );
  void changePhone( char *str );
  void changeDial( char *str );
  void changeDialSuff( char *str );
  void changeFlags( char *str );
  void changeRexxDial( char *str );
  void changeWait( int wt );
  void changeBaud( int wt );
  void changeFlavour( Flavour wt );
  void changeMaxIn( short wt );
  void changeMaxOut( short wt );
  void changeMaxTry( int wt );
  void changeCost( int wt );
  void changeNoDZAenabled( int wt );
  void changeNoZAPenabled( int wt );
  void changeNoHydraEnabled( int wt );
  void changeAfter( Schedule *wt );
  void changeBefore( Schedule *wt );
  void changeDoit( Schedule *wt );
  void changeChain( int ch );
  void changePriority( int prio );
  void changeAlias( char *str );
};

class NodeConfigList : public Collection
{
public:
  NodeConfigList();

  NodeConfig* Find( NodeConfig& item, LockMode lock=LockModeNeed ) const;
  NodeConfig& operator [] ( const int pos ) const;

  void postProcess();
  NodeConfig *FindAlias( const char *alias );
};

class NodeConfigIterator : public CollectionIterator
{
public:
  NodeConfigIterator( NodeConfigList& ini );
  
  NodeConfig& operator () () const;
};

extern NodeConfigList *systemList;
extern NodeConfigList *maskList;
extern FIDOAddress *defaultAddress;

int checkAddress( char *addr, uShort& zone, uShort& net,
	                 uShort& node, uShort& point,
		         char **domain, FIDOAddress *def=defaultAddress );

int splitLongName( char *box, uShort& zone, uShort& net,
		          uShort& node, uShort& point,
		          char **domain, Flavour& fl, LongNameType& type );

void moveNodeFiles( FIDOAddress *adr, char *path );

class FLO
{
  FILE *xlo;
  char ok;
  char *path;
  
protected:
  void purge();

public:
  FLO( char *_path );
  ~FLO();

  int process( int, int (*func)( char *, char, void * ), void * );
};

void postProcessAttach( char flag, char *test );

#endif
